#####################################################################################
# Commented code example for calculating an "Ensemble of Small Modells" (ESM) in R
# 
# This script compares predictions of standard Maxent models with ESMs composed of   
# bivariate Maxent models. Data include environmental data for presences of five 
# species and 10000 pseudo-absences used in Breiner et al.
#
# Due to restrictions of the data owner species are named as 'spec1' .. 'spec5'. 
#
# Breiner et al. found for rare species ESMs to be on average superior to Maxent
# when compared by cross-validated AUC or BOYCE index. While BOYCE generally showed 
# better performance of ESMs, AUC was in a few cases indifferent or decreased slightly
# (see spec5 of this example).
#
# Compared to Breiner et al. this is a simplified comparison using 50% of data 
# (random selection) for training and the other half of the data for testing 
# the models. This procedure is repeated 10 times for the selected species.
#
# NOTE: ESM functions are provided as part of the R package ecospat (Broennimann et al. 2016)
#
# REFERENCES:
# Breiner F.T., Guisan A., Bergamini A., Nobis M.P. (2015):
#  Overcoming limitations of modelling rare species by using ensembles of small models. Methods Ecol. Evol. 6: 1210-1218.
# Broenniman O.; Di Cola V.; Guisan A., 2016:
#  ecospat: Spatial Ecology Miscellaneous Methods. R package version 2.0. Available from World Wide Web <http://cran.r-project.org/package=ecospat> 
# Lomba, A., Pellissier, L., Randin, C.F., Vicente, J., Moreira, F., Honrado, J. & Guisan, A. (2010):
#  Overcoming the rare species modelling paradox: A novel hierarchical framework applied to an Iberian endemic plant. Biological Conservation, 143, 2647-2657.
#
# Frank Breiner, Swiss Federal Research Institute WSL, frank.breiner@wsl.ch, May 2016
#####################################################################################

# used packages

  library(dismo)
  library(ecospat)

# Maxent is used as model technique; Maxent can be downloaded form: 
# http://www.cs.princeton.edu/~schapire/maxent/
# The maxent.jar file needs to be copied to following folder:
# paste(system.file(package="dismo"), "/java", sep='')
# If there is an error "rJava cannot be loaded", specify the folder where JAVA is 
# installed: Sys.setenv(JAVA_HOME='path-to-your-java-installation')
# - and make sure the right 32bit or 64bit Java version is installed for your R.

  jar <- paste(system.file(package="dismo"), "/java/maxent.jar", sep='')
  if (!file.exists(jar)) { stop('maxent.jar not available')}

# read data and build datasets for modelling

  data <- read.table('C:/Users/Breiner/Desktop/ESM/EMS-data-5-species.csv', header=T, sep=',')
  dim(data) # 10212    18

  # environmental data: 11 predictos
  # "caco3","sfro","prec","pday","srad","tave","eness","nness","slp","topo","twi"
  # selection of 6 predictors for better runtime behavior
  env  <- data[, colnames(data) %in% c("caco3","prec","tave","nness","slp","topo")]

  # selection vector for 10'000 pseudo-absences
  sel.pa <- data$pa==1  
  sum(sel.pa)

  # species selection: spec1, spec2, spec3, spec4, or spec5
  sel.occ <- data$spec1==1  # choice of species: data$spec2 etc.
  sum(sel.occ)

  # species specific environmental data & response for modelling
  env.occ <- env[sel.occ,]
  env.bg  <- env[sel.pa ,]
  env  <- as.data.frame(rbind(env.occ, env.bg))
  resp <- c(rep(1,nrow(env.occ)),rep(0,nrow(env.bg)))
  
# =================================================================================
# 10 runs using a random 50:50 split of data for training and testing ESM vs Maxent
# =================================================================================

  eval.test.st  <-  NULL    # for evaluation results of Maxent standard
  eval.test.esm <-  NULL    # for evaluation results of ESM

  for (i in 1:10) {  

    # split data for training and testing (50:50)
    mykfold <- kfold(resp, k=2, by=resp)
    sel.train  <- mykfold == 1
    sel.test   <- mykfold != 1
    resp.train <- resp[sel.train]
    resp.test  <- resp[sel.test]

    # standard Maxent
    st <- maxent(x=env[sel.train,], p=resp[sel.train], args=c("-P","nohinge", "nothreshold", "noproduct"))                
    pred.st <- predict(st, env)    
    auc.test <- evaluate(p=pred.st[sel.test ][resp.test ==1], a=pred.st[sel.test ][resp.test ==0])@auc
    boyce.test <- ecospat.boyce(pred.st[sel.test], pred.st[sel.test][resp.test==1], PEplot=F)$Spearman.cor
    eval.test.st <- rbind(eval.test.st, cbind(auc=auc.test,boyce=boyce.test))

    # build ESM as weighted average of bivariate Maxent models
    weight.vec <- vector()                 # vector with weights of single bivariate models
    pred.biva.sum <- rep(0, length(resp))  # for ensemble prediction of bivariate models 
     
    for(m in 1:(dim(env)[2]-1)){           # build all bivariate predictor combinations
      for(n in (m+1):dim(env)[2]){

        # calibrate bivariate maxent model
        me.biva <- maxent(x=env[sel.train,c(m,n)], p=resp.train, args=c("-P","nohinge", "nothreshold", "noproduct"))  
         
        # prediction
        pred.biva <- predict(me.biva, env)
         
        # evaluate bivariate model and calculate its weight for ensemble prediction
        eval.train.biva <- evaluate(p=pred.biva[sel.train][resp.train==1], a=pred.biva[sel.train][resp.train==0])
        weight <- eval.train.biva@auc*2-1  # use Somers' D as weighting factor
        if(weight < 0) {weight <- 0}
        weight.vec <- c(weight.vec, weight)
         
        # build weighted sum of bivariate predictions
        pred.biva.sum <- pred.biva.sum + (pred.biva * weight)
         
      }
    }
     
    # calculate ESM prediction
    pred.esm <- pred.biva.sum/sum(weight.vec)
     
    # evaluate ESM
    auc.test <- evaluate(p=pred.esm[sel.test ][resp.test ==1], a=pred.esm[sel.test ][resp.test ==0])@auc
    boyce.test <- ecospat.boyce(pred.esm[sel.test],pred.esm[sel.test][resp.test==1], PEplot=F)$Spearman.cor     
    eval.test.esm <- rbind(eval.test.esm , cbind(auc=auc.test,boyce=boyce.test))

    # show result of a single split
    message('split ',i, ' of 10:',
      '  AUC.maxent = '  , round(eval.test.st[i,"auc"],3),
      ', AUC.esm = '     , round(eval.test.esm[i,"auc"],3),
      ', BOYCE.maxent = ', round(eval.test.st[i,"boyce"],3),
      ', BOYCE.esm = '   , round(eval.test.esm[i,"boyce"],3), sep=''
    )
  }

# show result of all 10 random splits

   par(mfrow=c(1,4))
   boxplot(eval.test.st[,"auc"],eval.test.esm[,"auc"],names=c("standard","ESM"),main="AUC")
   boxplot(eval.test.st[,"boyce"],eval.test.esm[,"boyce"],names=c("standard","ESM"),main="Boyce")
   hist(eval.test.esm[,"auc"]-eval.test.st[,"auc"], main='delta AUC', col='lightgrey', xlab='auc.esm - auc.standard')
   lines(c(0,0),c(0,99), col='red', lw=2)
   hist(eval.test.esm[,"boyce"]-eval.test.st[,"boyce"], main='delta Boyce', col='lightgrey', xlab='boyce.esm - boyce.standard')
   lines(c(0,0),c(0,99), col='red', lw=2)

# EOF (end of file)